@@ -17,9 +17,8 @@
 -- A Yesod authentication plugin designed to look users up in a Persistent
 -- database where the hash of their password is stored.
 --
--- __Release 1.5 is the first of two which remove compatibility with old__
--- __(pre 1.3) databases.  Some other deprecated functionality is removed__
--- __too.  Please see__
+-- __Releases 1.6 finishes the process of removing compatibility with old__
+-- __(pre 1.3) databases.  Please see__
 -- __<https://github.com/paul-rouse/yesod-auth-hashdb/blob/master/Upgrading.md>__
 --
 -- To use this in a Yesod application, the foundation data type must be an
@@ -105,6 +104,52 @@
 -- included in the widget to add it - see @defaultForm@ in the source
 -- code of this module for an example.
 --
+-- == JSON Interface
+--
+-- This plugin provides sufficient tools to build a complete JSON-based
+-- authentication flow.  We assume that a design goal is to avoid URLs
+-- being built into the client, so all of the URLs needed are passed in
+-- JSON data.
+--
+-- To start the process, Yesod's defaultErrorHandler produces a JSON
+-- response if the HTTP Accept header gives \"application/json\"
+-- precedence over HTML.  For a NotAuthenticated error, the status is
+-- 401 and the response contains the URL to use for authentication: this
+-- is the route which will be handled by the loginHandler method of the
+-- YesodAuth instance, which normally returns a login form.
+--
+-- Leaving the loginHandler aside for a moment, the final step - supported
+-- by this plugin since version 1.6 - is to POST the credentials for
+-- authentication in a JSON object.  This object must include the
+-- properties "username" and "password".  In the HTML case this would be
+-- the form submission, but here we want to use JSON instead.
+--
+-- In a JSON interface, the purpose of the loginHandler is to tell the
+-- client the URL for submitting the credentials.  This requires a
+-- custom loginHandler, since the default one generates HTML only.
+-- It can find the correct URL by using the 'submitRouteHashDB'
+-- function defined in this module.
+--
+-- Writing the loginHandler is made a little messy by the fact that its
+-- type allows only HTML content.  A work-around is to send JSON as a
+-- short-circuit response, but we still make the choice using selectRep
+-- so as to get its matching of content types.  Here is an example which
+-- is geared around using HashDB on its own, supporting both JSON and HTML
+-- clients:
+--
+-- > instance YesodAuth App where
+-- >    ....
+-- >    loginHandler = do
+-- >         submission <- submitRouteHashDB
+-- >         render <- lift getUrlRender
+-- >         typedContent@(TypedContent ct _) <- selectRep $ do
+-- >             provideRepType typeHtml $ return emptyContent
+-- >                            -- Dummy: the real Html version is at the end
+-- >             provideJson $ object [("loginUrl", toJSON $ render submission)]
+-- >         when (ct == typeJson) $
+-- >             sendResponse typedContent   -- Short-circuit JSON response
+-- >         defaultLoginHandler             -- Html response
+--
 -------------------------------------------------------------------------------
 module Yesod.Auth.HashDB
     ( HashDBUser(..)
@@ -117,18 +162,19 @@ module Yesod.Auth.HashDB
     , validateUser
     , authHashDB
     , authHashDBWithForm
+    , submitRouteHashDB
     ) where
 
 
 #if __GLASGOW_HASKELL__ < 710
-import           Control.Applicative         ((<$>), (<*>))
+import           Control.Applicative         ((<$>), (<*>), pure)
 #endif
-import qualified Crypto.Hash           as CH (SHA1, Digest, hash)
 import           Crypto.PasswordStore        (makePassword, strengthenPassword,
                                               verifyPassword, passwordStrength)
+import           Data.Aeson                  ((.:?))
 import qualified Data.ByteString.Char8 as BS (pack, unpack)
 import           Data.Maybe                  (fromMaybe)
-import           Data.Text                   (Text, pack, unpack, append)
+import           Data.Text                   (Text, pack, unpack)
 import           Yesod.Auth
 import qualified Yesod.Auth.Message    as Msg
 import           Yesod.Core
@@ -146,36 +192,23 @@ defaultStrength :: Int
 defaultStrength = 17
 
 -- | The type representing user information stored in the database should
---   be an instance of this class.  It just provides the getters and setters
+--   be an instance of this class.  It just provides the getter and setter
 --   used by the functions in this module.
 class HashDBUser user where
-    -- | Retrieve password hash from user data
+    -- | Getter used by 'validatePass' and 'upgradePasswordHash' to
+    --   retrieve the password hash from user data
+    --
     userPasswordHash :: user -> Maybe Text
-    -- | Retrieve salt for password from user data.  This is needed only for
-    --   compatibility with old database entries, which contain the salt
-    --   as a separate field.  New implementations do not require a separate
-    --   salt field in the user data, and should leave this as the default.
-    userPasswordSalt :: user -> Maybe Text
-    userPasswordSalt _ = Just ""
 
-    -- | Callback for 'setPassword' and 'upgradePasswordHash'.  Produces a
+    -- | Setter used by 'setPassword' and 'upgradePasswordHash'.  Produces a
     --   version of the user data with the hash set to the new value.
     --
     setPasswordHash :: Text   -- ^ Password hash
                        -> user -> user
 
     {-# MINIMAL userPasswordHash, setPasswordHash #-}
-{-# DEPRECATED userPasswordSalt "Verification against old data containing a separate salt field will be removed in version 1.6" #-}
 
 
--- | Calculate salted hash using SHA1.  Retained for compatibility with
---   hashes in existing databases, but will not be used for new passwords.
-saltedHash :: Text              -- ^ Salt
-           -> Text              -- ^ Password
-           -> Text
-saltedHash salt pw =
-    pack $ show (CH.hash $ BS.pack $ unpack $ append salt pw :: CH.Digest CH.SHA1)
-
 -- | Calculate a new-style password hash using "Crypto.PasswordStore".
 passwordHash :: MonadIO m => Int -> Text -> m Text
 passwordHash strength pwd = do
@@ -200,11 +233,6 @@ setPassword :: (MonadIO m, HashDBUser user) => Text -> user -> m user
 setPassword = setPasswordStrength defaultStrength
 
 -- | Validate a plaintext password against the hash in the user data structure.
---   This function retains compatibility with user data produced by old
---   versions of this module (prior to 1.3), although the hashes are less
---   secure and should be upgraded as soon as possible.  They can be
---   upgraded using 'upgradePasswordHash', or by insisting that users set
---   new passwords.
 --
 --   The result distinguishes two types of validation failure, which may
 --   be useful in an application which supports multiple authentication
@@ -213,27 +241,24 @@ setPassword = setPasswordStrength defaultStrength
 --   * Just False - the user has a password set up, but the given one does
 --     not match it
 --
---   * Nothing - the user does not have a password (the hash is Nothing)
+--   * Nothing - the user does not have a password ('userPasswordHash' returns
+--     Nothing)
 --
 --   Since 1.4.1
 --
 validatePass :: HashDBUser u => u -> Text -> Maybe Bool
 validatePass user passwd = do
     hash <- userPasswordHash user
-    salt <- userPasswordSalt user
-    -- NB plaintext password characters are truncated to 8 bits here, and also
-    -- in saltedHash and passwordHash above (the hash and old salt are already
-    -- 8 bit).  This is for historical compatibility, but in practice it is
+    -- NB plaintext password characters are truncated to 8 bits here,
+    -- and also in passwordHash above (the hash is already 8 bit).
+    -- This is for historical compatibility, but in practice it is
     -- unlikely to reduce the entropy of most users' alphabets by much.
     let hash' = BS.pack $ unpack hash
-        passwd' = BS.pack $ unpack $ if salt == "" then passwd
-                                     else
-                                         -- Extra layer for an upgraded old hash
-                                         saltedHash salt passwd
+        passwd' = BS.pack $ unpack passwd
     if passwordStrength hash' > 0
-        -- Will give >0 for new-style hash, else fall back
+        -- Will give >0 for valid hash format, else treat as if wrong password
         then return $ verifyPassword passwd' hash'
-        else return $ hash == saltedHash salt passwd
+        else return False
 
 -- | Upgrade existing user credentials to a stronger hash.  The existing
 --   hash will have been produced from a weaker setting in the current
@@ -246,25 +271,21 @@ validatePass user passwd = do
 --   but may have been upgraded using earlier versions of this function.
 --
 --   Returns Nothing if the user has no password (ie if 'userPasswordHash' u
---   is 'Nothing'), if the password hash is not in the new format, or if the
---   user has a non-empty salt field resulting from the old-style hashing
---   algorithm.
+--   is 'Nothing') or if the password hash is not in the correct format.
+--
 upgradePasswordHash :: (MonadIO m, HashDBUser user) => Int -> user -> m (Maybe user)
 upgradePasswordHash strength u = do
     let old = userPasswordHash u
-        oldSalt = fromMaybe "" $ userPasswordSalt u
     case old of
         Just oldHash -> do
             let oldHash' = BS.pack $ unpack oldHash
             if passwordStrength oldHash' > 0
               then
-                -- Already a new-style hash, so only strengthen it as needed
+                -- Valid hash format, so strengthen it as needed
                 let newHash = pack $ BS.unpack $ strengthenPassword oldHash' strength
-                in if oldSalt == ""
-                   then return $ Just $ setPasswordHash newHash u
-                   else return Nothing   -- Can't handle non-empty salt
+                in return $ Just $ setPasswordHash newHash u
               else do
-                -- Old-style hash: cannot upgrade
+                -- Invalid hash format (perhaps from old version of this module)
                 return Nothing
         Nothing -> return Nothing
 
@@ -288,6 +309,15 @@ type HashDBPersist master user =
     , PersistEntity user
     )
 
+-- Internal data type for receiving JSON encoded username and password
+data UserPass = UserPass (Maybe Text) (Maybe Text)
+
+instance FromJSON UserPass where
+    parseJSON (Object v) = UserPass <$>
+                           v .:? "username" <*>
+                           v .:? "password"
+    parseJSON _          = pure $ UserPass Nothing Nothing
+
 -- | Given a user ID and password in plaintext, validate them against
 --   the database values.  This function simply looks up the user id in the
 --   database and calls 'validatePass' to do the work.
@@ -308,13 +338,21 @@ login = PluginR "hashdb" ["login"]
 
 -- | Handle the login form. First parameter is function which maps
 --   username (whatever it might be) to unique user ID.
+--
+--   Since version 1.6, the data may be submitted as a JSON object.
+--   See the \"JSON Interface\" section above for more details.
 postLoginR :: HashDBPersist site user =>
               (Text -> Maybe (Unique user))
            -> HandlerT Auth (HandlerT site IO) TypedContent
 postLoginR uniq = do
-    (mu,mp) <- lift $ runInputPost $ (,)
-        <$> iopt textField "username"
-        <*> iopt textField "password"
+    ct <- lookupHeader "Content-Type"
+    let jsonContent = ((== "application/json") . simpleContentType) <$> ct
+    UserPass mu mp <-
+        case jsonContent of
+          Just True -> requireJsonBody
+          _         -> lift $ runInputPost $ UserPass
+                       <$> iopt textField "username"
+                       <*> iopt textField "password"
 
     isValid <- lift $ fromMaybe (return False) 
                  (validateUser <$> (uniq =<< mu) <*> mp)
@@ -386,3 +424,17 @@ defaultForm loginRoute = do
             }
 
     |]
+
+
+-- | The route, in the parent site, to which the username and password
+--   should be sent in order to log in.  This function is particularly
+--   useful in constructing a 'loginHandler' function which provides a
+--   JSON response.  See the \"JSON Interface\" section above for more
+--   details.
+--
+--   Since 1.6
+submitRouteHashDB :: YesodAuth site =>
+                  HandlerT Auth (HandlerT site IO) (Route site)
+submitRouteHashDB = do
+    toParent <- getRouteToParent
+    return $ toParent login
